@@ -19,15 +19,28 @@
 
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
-
 """A simple colormap module."""
 
+import contextlib
+import os
+from io import StringIO
+from typing import Optional
 import warnings
+from numbers import Number
 
 import numpy as np
 from trollimage.colorspaces import rgb2hcl, hcl2rgb
 
 
+@contextlib.contextmanager
+def _file_or_stringio(filename_or_none):
+    if filename_or_none is None:
+        yield StringIO()
+    else:
+        with open(filename_or_none, "w") as file_obj:
+            yield file_obj
+
+
 def colorize(arr, colors, values):
     """Colorize a monochromatic array *arr*, based *colors* given for *values*.
 
@@ -71,10 +84,16 @@ def _colorize(arr, colors, values):
 
 
 def _interpolate_rgb_colors(arr, colors, values):
+    interp_xp_coords = np.array(values)
     hcl_colors = _convert_rgb_list_to_hcl(colors)
+    interp_y_coords = np.array(hcl_colors)
+    if values[0] > values[-1]:
+        # monotonically decreasing
+        interp_xp_coords = interp_xp_coords[::-1]
+        interp_y_coords = interp_y_coords[::-1]
     channels = [np.interp(arr,
-                          np.array(values),
-                          np.array(hcl_colors)[:, i])
+                          interp_xp_coords,
+                          interp_y_coords[:, i])
                 for i in range(3)]
     channels = list(hcl2rgb(*channels))
     return channels
@@ -100,11 +119,7 @@ def _interpolate_alpha(arr, colors, values):
 
 def _mask_channels(channels, arr):
     """Mask the channels if arr is a masked array."""
-    try:
-        channels = [_mask_array(channel, arr) for channel in channels]
-    except AttributeError:
-        pass
-    return channels
+    return [_mask_array(channel, arr) for channel in channels]
 
 
 def _mask_array(new_array, arr):
@@ -145,10 +160,17 @@ def _palettize(arr, values):
 
 
 def _digitize_array(arr, values):
-    new_arr = np.digitize(arr.ravel(),
-                          np.concatenate((values,
-                                          [max(np.nanmax(arr),
-                                               values.max()) + 1])))
+    if values[0] <= values[-1]:
+        # monotonic increasing values
+        outside_range_bin = max(np.nanmax(arr), values.max()) + 1
+        right = False
+    else:
+        # monotonic decreasing values
+        outside_range_bin = min(np.nanmin(arr), values.min()) - 1
+        right = True
+    bins = np.concatenate((values, [outside_range_bin]))
+
+    new_arr = np.digitize(arr.ravel(), bins, right=right)
     new_arr -= 1
     new_arr = new_arr.clip(min=0, max=len(values) - 1)
     return new_arr
@@ -163,7 +185,7 @@ class Colormap(object):
             aren't provided.
         values: One dimensional array-like of control points where
             each corresponding color is applied. Must be the same number of
-            elements as colors and must be monotonically increasing.
+            elements as colors and must be monotonic.
         colors: Two dimensional array-like of RGB or RGBA colors where each
             color is applied to a specific control point. Must be the same
             number of colors as control points (values). Colors should be
@@ -256,15 +278,22 @@ class Colormap(object):
         """Append colormap together."""
         old, other = self._normalize_color_arrays(self, other)
         values = np.concatenate((old.values, other.values))
-        if not (np.diff(values) >= 0).all():
+        if not self._monotonic_one_direction(values):
             raise ValueError("Merged colormap 'values' are not monotonically "
-                             "increasing or equal.")
+                             "increasing, monotonically decreasing, or equal.")
         colors = np.concatenate((old.colors, other.colors))
         return Colormap(
             values=values,
             colors=colors,
         )
 
+    @staticmethod
+    def _monotonic_one_direction(values):
+        delta = np.diff(values)
+        all_increasing = (delta >= 0).all()
+        all_decreasing = (delta <= 0).all()
+        return all_increasing or all_decreasing
+
     @staticmethod
     def _normalize_color_arrays(cmap1, cmap2):
         num_bands1 = cmap1.colors.shape[-1]
@@ -293,9 +322,9 @@ class Colormap(object):
     def set_range(self, min_val, max_val, inplace=True):
         """Set the range of the colormap to [*min_val*, *max_val*].
 
-        If min is greater than max then the Colormap's colors are reversed
-        before values are updated to the new range. This is done because
-        Colormap ``values`` must always be monotonically increasing.
+        The Colormap's values will match the range specified even if "min_val"
+        is greater than "max_val". To flip the order of the colors, use
+        :meth:`reversed`.
 
         Args:
             min_val (float): New minimum value for the control points in
@@ -306,14 +335,9 @@ class Colormap(object):
                 If False, return a new Colormap instance.
 
         """
-        if min_val > max_val:
-            cmap = self.reverse(inplace=inplace)
-            max_val, min_val = min_val, max_val
-        else:
-            cmap = self
-
-        values = (((cmap.values * 1.0 - cmap.values.min()) /
-                   (cmap.values.max() - cmap.values.min()))
+        cmap = self
+        values = (((cmap.values * 1.0 - cmap.values[0]) /
+                   (cmap.values[-1] - cmap.values[0]))
                   * (max_val - min_val) + min_val)
         if not inplace:
             return Colormap(
@@ -325,10 +349,157 @@ class Colormap(object):
         return cmap
 
     def to_rio(self):
-        """Convert the colormap to a rasterio colormap."""
+        """Convert the colormap to a rasterio colormap.
+
+        Note that rasterio requires color tables to have round integer value
+        control points. This method assumes that the range of this Colormap
+        is already in the desired output range and to avoid issues with
+        rasterio will round the values and convert them to unsigned integers.
+        """
         colors = (((self.colors * 1.0 - self.colors.min()) /
                    (self.colors.max() - self.colors.min())) * 255)
-        return dict(zip(self.values, tuple(map(tuple, colors))))
+        # rasterio doesn't allow non-integer colormap values
+        values = np.round(self.values).astype(np.uint)
+        return dict(zip(values, tuple(map(tuple, colors))))
+
+    def to_csv(
+            self,
+            filename: Optional[str] = None,
+            color_scale: Number = 255,
+    ) -> Optional[str]:
+        """Save Colormap to a comma-separated text file or string.
+
+        The CSV data will have 4 to 5 columns for each row where each
+        each row will contain the value (V), red (R), green (B), blue (B),
+        and if configured alpha (A).
+
+        The values will remain in whatever range is currently set on the
+        colormap. The colors of the colormap (assumed to be between 0 and 1)
+        will be multiplied by 255 to produce a traditional unsigned 8-bit
+        integer value.
+
+        Args:
+            filename: The filename of the CSV file to save to.
+                If not provided or None a string is returned with the contents.
+            color_scale: Scale colors by this factor before converting to a
+                CSV. Colors are stored in the Colormap in a 0 to 1 range.
+                Defaults to 255. If not equal to 1 values are converted to
+                integers too.
+
+        """
+        with _file_or_stringio(filename) as csv_file:
+            for value, color in zip(self.values, self.colors):
+                scaled_color = [x * color_scale for x in color]
+                if color_scale != 1.0:
+                    scaled_color = [int(x) for x in scaled_color]
+                csv_file.write(",".join(["{:0.6f}".format(value)] + [str(x) for x in scaled_color]) + "\n")
+        if isinstance(csv_file, StringIO):
+            return csv_file.getvalue()
+
+    @classmethod
+    def from_file(
+            cls,
+            filename_or_string: str,
+            colormap_mode: Optional[str] = None,
+            color_scale: Number = 255,
+    ):
+        """Create Colormap from a comma-separated or binary file of colormap data.
+
+        Args:
+            filename_or_string: Filename of a binary or CSV file or a
+                string version of the comma-separate data.
+            colormap_mode: Force the scheme of the colormap data (ex. RGBA).
+                See information below on other possible values and how they
+                are interpreted. By default this is determined based on the
+                number of columns in the data.
+            color_scale: The maximum possible color value in the colormap data
+                provided. For example, if the colors in the provided data were
+                8-bit unsigned integers this should be 255 (the default). This
+                value will be used to normalize the colors from 0 to 1.
+
+        Colormaps can be loaded from ``.npy``, ``.npz``, or comma-separated text
+        files. Numpy (npy/npz) files should be 2D arrays with rows for each color.
+        Comma-separated files should have a row for each color with each column
+        representing a single value/channel. A filename
+        ending with ``.npy`` or ``.npz`` is read as a numpy file with
+        :func:`numpy.load`. All other extensions are
+        read as a comma-separated file. For ``.npz`` files the data must be stored
+        as a positional list where the first element represents the colormap to
+        use. See :func:`numpy.savez` for more information. The filename should
+        be an absolute path for consistency.
+
+        The colormap is interpreted as 1 of 4 different "colormap modes":
+        ``RGB``, ``RGBA``, ``VRGB``, or ``VRGBA``. The
+        colormap mode can be forced with the ``colormap_mode`` keyword
+        argument. If it is not provided then a default will be chosen
+        based on the number of columns in the array (3: RGB, 4: VRGB, 5: VRGBA).
+
+        The "V" in the possible colormap modes represents the control value of
+        where that color should be applied. If "V" is not provided in the colormap
+        data it defaults to the row index in the colormap array (0, 1, 2, ...)
+        divided by the total number of colors to produce a number between 0 and 1.
+        See the "Set Range" section below for more information.
+        The remaining elements in the colormap array represent the Red (R),
+        Green (G), and Blue (B) color to be mapped to.
+
+        See the "Color Scale" section below for more information on the value
+        range of provided numbers.
+
+        **Color Scale**
+
+        By default colors are expected to be in a 0-255 range. This
+        can be overridden by specifying ``color_scale`` keyword argument.
+        A common alternative to 255 is ``1`` to specify floating
+        point numbers between 0 and 1. The resulting Colormap uses the normalized
+        color values (0-1).
+
+        """
+        if not os.path.isfile(filename_or_string):
+            filename_or_string = StringIO(filename_or_string)
+        values, colors = _get_values_colors_from_file(filename_or_string, colormap_mode, color_scale)
+        return cls(values=values, colors=colors)
+
+
+def _get_values_colors_from_file(filename, colormap_mode, color_scale):
+    data = _read_colormap_data_from_file(filename)
+    cols = data.shape[1]
+    default_modes = {
+        3: 'RGB',
+        4: 'VRGB',
+        5: 'VRGBA'
+    }
+    default_mode = default_modes.get(cols)
+    if colormap_mode is None:
+        colormap_mode = default_mode
+    if colormap_mode is None or len(colormap_mode) != cols:
+        raise ValueError(
+            "Unexpected colormap shape for mode '{}'".format(colormap_mode))
+    rows = data.shape[0]
+    if colormap_mode[0] == 'V':
+        colors = data[:, 1:]
+        if color_scale != 1:
+            colors = data[:, 1:] / float(color_scale)
+        values = data[:, 0]
+    else:
+        colors = data
+        if color_scale != 1:
+            colors = colors / float(color_scale)
+        values = np.arange(rows) / float(rows - 1)
+    return values, colors
+
+
+def _read_colormap_data_from_file(filename_or_file_obj):
+    if isinstance(filename_or_file_obj, str):
+        ext = os.path.splitext(filename_or_file_obj)[1]
+        if ext in (".npy", ".npz"):
+            file_content = np.load(filename_or_file_obj)
+            if ext == ".npz":
+                # .npz is a collection
+                # assume position list-like and get the first element
+                file_content = file_content["arr_0"]
+            return file_content
+    # CSV file or file-like object of CSV string data
+    return np.loadtxt(filename_or_file_obj, delimiter=",")
 
 
 # matlab jet "#00007F", "blue", "#007FFF", "cyan", "#7FFF7F", "yellow",
